// File         : winappstdio.cpp
// Author       : Brian Boylston
// Purpose      : enables win32 applications to use standard C and C++ io
// Created      : 12/21/1998
// Last Modified: 12/21/1998


// includes
//

#include <io.h>           // low level i/o routines
#include <stdio.h>        // C stdio
#include <errno.h>        // C runtime error codes
#include <windows.h>      // win32

#include "winappstdio.h"  // our header


// private constants
//

#define STDIN_C_FD   0  // C runtime file descriptor for stdin
#define STDOUT_C_FD  1  // C runtime file descriptor for stdout
#define STDERR_C_FD  2  // C runtime file descriptor for stderr


// private types
//

typedef struct _win_app_stdio_t
{
	HANDLE   hConStdIn;   // win32 console stdin handle
	HANDLE   hConStdOut;  // win32 console stdout handle
	HANDLE   hConStdErr;  // win32 console stderr handle
	int      fdStdIn;     // C runtime fd for win32 stdin
	int      fdStdOut;    // C runtime fd for win32 stdout
	int      fdStdErr;    // C runtime fd for win32 stderr
	FILE     oldstdin;    // C runtime FILE struct for original stdin
	FILE     oldstdout;   // C runtime FILE struct for original stdout
	FILE     oldstderr;   // C runtime FILE struct for original stderr
	FILE*    pstdin;      // new C runtime stdin
	FILE*    pstdout;     // new C runtime stdout
	FILE*    pstderr;     // new C runtime stderr
} win_app_stdio_t;


// private globals
//

win_app_stdio_t*   gpInfo;  // private global handle info


// implementation
//

int OpenWinAppConsole( const char* pszConsoleTitle )
{
	DWORD   dwError;  // error returned from GetLastError()

	// by default, success
	dwError = 0;

	// close any previous console window
	if( !CloseWinAppConsole() )
		// error closing previous console window
		return 0;

	// allocate new global storage
	gpInfo = new win_app_stdio_t;
	if( !gpInfo ) {
		// error allocated memory
		SetLastError( ERROR_NOT_ENOUGH_MEMORY );
		return 0;
	}
	memset( (void*)gpInfo, 0, sizeof(win_app_stdio_t) );

	// open a console window
	if( !AllocConsole() ) {
		// unable to open a console window
		dwError = GetLastError();
		// cleanup
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}

	// *** HACK
	//
//	HANDLE   hFileRedirect;
//	hFileRedirect = CreateFile( "out.txt", GENERIC_WRITE, 0, NULL,
//		                        CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL );

	// get console window std handles
	gpInfo->hConStdIn = GetStdHandle( STD_INPUT_HANDLE );
	if( gpInfo->hConStdIn == INVALID_HANDLE_VALUE ) {
		// unable to get console window stdin handle
		dwError = GetLastError();
		// cleanup
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	gpInfo->hConStdOut = GetStdHandle( STD_OUTPUT_HANDLE );
	if( gpInfo->hConStdOut == INVALID_HANDLE_VALUE ) {
		// unable to get console window stdout handle
		dwError = GetLastError();
		// cleanup
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	gpInfo->hConStdErr = GetStdHandle( STD_ERROR_HANDLE );
	if( gpInfo->hConStdErr == INVALID_HANDLE_VALUE ) {
		// unable to get console window stderr handle
		dwError = GetLastError();
		// cleanup
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}

	// open new C runtime handles
	gpInfo->fdStdIn = _open_osfhandle( (long)(gpInfo->hConStdIn), 0 );
	if( gpInfo->fdStdIn == -1 ) {
		// error opening new C runtime handle for stdin
		dwError = GetLastError();
		// cleanup
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	gpInfo->fdStdOut = _open_osfhandle( (long)(gpInfo->hConStdOut), 0 );
	if( gpInfo->fdStdOut == -1 ) {
		// error opening new C runtime handle for stdout
		dwError = GetLastError();
		// cleanup
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	gpInfo->fdStdErr = _open_osfhandle( (long)(gpInfo->hConStdErr), 0 );
	if( gpInfo->fdStdErr == -1 ) {
		// error opening new C runtime handle for stderr
		dwError = GetLastError();
		// cleanup
		_close( gpInfo->fdStdOut );
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}

	// from here, do or die
	fclose( stdin );
	fclose( stdout );
	fclose( stderr );

	// connect new C runtime fds to standard C runtime fds
	if( _dup2(gpInfo->fdStdIn, STDIN_C_FD) == -1 ) {
		// error connecting new stdin handle to stdin fd
		if( errno == EBADF )
			dwError = ERROR_INVALID_HANDLE;
		else
			dwError = ERROR_TOO_MANY_OPEN_FILES;
		// cleanup
		_close( gpInfo->fdStdErr );
		_close( gpInfo->fdStdOut );
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	if( _dup2(gpInfo->fdStdOut, STDOUT_C_FD) == -1 ) {
		// error connecting new stdout handle to stdin fd
		if( errno == EBADF )
			dwError = ERROR_INVALID_HANDLE;
		else
			dwError = ERROR_TOO_MANY_OPEN_FILES;
		// cleanup
		_close( STDIN_C_FD );
		_close( gpInfo->fdStdErr );
		_close( gpInfo->fdStdOut );
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	if( _dup2(gpInfo->fdStdErr, STDERR_C_FD) == -1 ) {
		// error connecting new stdin handle to stderr fd
		if( errno == EBADF )
			dwError = ERROR_INVALID_HANDLE;
		else
			dwError = ERROR_TOO_MANY_OPEN_FILES;
		// cleanup
		_close( STDOUT_C_FD );
		_close( STDIN_C_FD );
		_close( gpInfo->fdStdErr );
		_close( gpInfo->fdStdOut );
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}

	// create new streams for stdin, stdout, and stderr
	gpInfo->pstdin = _fdopen( STDIN_C_FD, "rt" );
	if( !gpInfo->pstdin ) {
		// error creating stream for stdin
		dwError = GetLastError();
		// cleanup
		_close( STDERR_C_FD );
		_close( STDOUT_C_FD );
		_close( STDIN_C_FD );
		_close( gpInfo->fdStdErr );
		_close( gpInfo->fdStdOut );
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	memcpy( (void*)&(gpInfo->oldstdin), (void*)stdin, sizeof(FILE) );
	memcpy( (void*)stdin, (void*)(gpInfo->pstdin), sizeof(FILE) );
	gpInfo->pstdout = _fdopen( STDOUT_C_FD, "wt" );
	if( !gpInfo->pstdout ) {
		// error creating stream for stdout
		dwError = GetLastError();
		// cleanup
		_close( STDERR_C_FD );
		_close( STDOUT_C_FD );
		memcpy( (void*)stdin, (void*)&(gpInfo->oldstdin), sizeof(FILE) );
		fclose( gpInfo->pstdin );
		gpInfo->pstdin = NULL;
		_close( gpInfo->fdStdErr );
		_close( gpInfo->fdStdOut );
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	memcpy( (void*)&(gpInfo->oldstdout), (void*)stdout, sizeof(FILE) );
	memcpy( (void*)stdout, (void*)(gpInfo->pstdout), sizeof(FILE) );
	gpInfo->pstderr = _fdopen( STDERR_C_FD, "wt" );
	if( !gpInfo->pstderr ) {
		// error creating stream for stderr
		dwError = GetLastError();
		// cleanup
		_close( STDERR_C_FD );
		memcpy( (void*)stdout, (void*)&(gpInfo->oldstdout), sizeof(FILE) );
		fclose( gpInfo->pstdout );
		gpInfo->pstdout = NULL;
		memcpy( (void*)stdin, (void*)&(gpInfo->oldstdin), sizeof(FILE) );
		fclose( gpInfo->pstdin );
		gpInfo->pstdin = NULL;
		_close( gpInfo->fdStdErr );
		_close( gpInfo->fdStdOut );
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( hFileRedirect );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
		// return with error
		SetLastError( dwError );
		return 0;
	}
	memcpy( (void*)&(gpInfo->oldstderr), (void*)stderr, sizeof(FILE) );
	memcpy( (void*)stderr, (void*)(gpInfo->pstderr), sizeof(FILE) );

	// set console window title
	if( pszConsoleTitle ) {
		// use user-specified title
		if( !SetConsoleTitle(pszConsoleTitle) ) {
			// error setting console title
			dwError = GetLastError();
			// return error via GetLastError(), but
			// otherwise succeed
		}
	}

	// done
	SetLastError( dwError );
	return 1;

}  // OpenWinAppConsole()

int CloseWinAppConsole()
{
	// shutdown stdio
	if( gpInfo ) {
		memcpy( (void*)stderr, (void*)&(gpInfo->oldstderr), sizeof(FILE) );
		fclose( gpInfo->pstderr );
		gpInfo->pstderr = NULL;
		memcpy( (void*)stdout, (void*)&(gpInfo->oldstdout), sizeof(FILE) );
		fclose( gpInfo->pstdout );
		gpInfo->pstdout = NULL;
		memcpy( (void*)stdin, (void*)&(gpInfo->oldstdin), sizeof(FILE) );
		fclose( gpInfo->pstdin );
		gpInfo->pstdin = NULL;
		_close( gpInfo->fdStdErr );
		_close( gpInfo->fdStdOut );
		_close( gpInfo->fdStdIn );
		// *** HACK
//		CloseHandle( gpInfo->hConStdOut );
		FreeConsole();
		delete gpInfo;
		gpInfo = NULL;
	}

	// done
	return 1;

}  // CloseWinAppConsole()
